/*
 * Copyright (C) 1999 by   XCIN TEAM
 * Copyright (C) 2004 by   Leon Ho <llch@redhat.com>
 *                         Lawrence Lim <llim@redhat.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "xcin.h"


char *class_names[] = 
{
    "com.redhat.xcin.Panel",
    "com.redhat.xcin.Aux"
};


/*
 * define if_methods_t
 */
if_methods_t xcin_method =
{
    if_xcin_OpenIF,
    if_xcin_CloseIF,
    if_xcin_GetIFValue,
    if_xcin_SetIFValue,
    if_xcin_OpenDesktop,
    if_xcin_CloseDesktop,
    if_xcin_CreateSC,
    if_xcin_DestroySC,
    if_xcin_GetSCValue,
    if_xcin_SetSCValue,
    if_xcin_ResetSC,
    if_xcin_SetSCFocus,
    if_xcin_UnsetSCFocus,
    if_xcin_SendEvent
};

/*
 * define Human Readable Name as UTFCHAR for LE in Unicode
 * 'fun2 ti3 pin1 yin1 mo2 kuai4' | traditional chinese pin yin module |
 */
// NOTE: Yet to see this appearing anywhere in the application
UTFCHAR lename_string[] =
{
    0x7e41,	
    0x4f53,
    0x62fc,
    0x97f3,
    0x6a21,
    0x5757,
    0x0
};

/*
 * define Human Readable Name as UTFCHAR for language
 */
UTFCHAR zhhrn_string[] =
{
    0x7e41,
    0x4f53,
    0x62fc,
    0x97f3,	
    0x0
};

static IMLEName lename[] =
{
    "xcin",	        /* LE id */
    lename_string	/* HRN 	 */
};

static IMLocale locales[] =
{
	{
		"zh_TW",	    /* locale id */
		zhhrn_string	/* HRN	     */
	},
	NULL
};

UTFCHAR aux_name_panel[] = { 0x62fc, 0x97f3, 0 }; /* pin yin */

IMObjectDescriptorStruct* objects = NULL;

typedef struct
{
    UTFCHAR name[10];
    char file[30];
    int n_preedit;
} IMEntry;

static IMEntry imentries[] =
{
    { { 0x5009, 0x9821, 0 },                 "cj", 1 },
    { { 0x7c21, 0x6613, 0 },                 "simplex", 1 },
    { { 0x6ce8, 0x97f3, 0 },                 "phone", 1 },
    { { 0x62fc, 0x97f3, 0 },                 "pinyin", 1 },
    { { 0x7121, 0x8abf, 0x62fc, 0x97f3, 0 }, "pinyin0", 1 },
    { { 0x7cb5, 0x62fc, 0 },                 "jyutping0", 1 },
    { { 0x884c, 0x5217, 0 },                 "array30", 2 },
    { { 0x8f15, 0x9b06, 0 },                 "ez", 1 },
    { { 0x96fb, 0x5831, 0 },                 "telecode", 1 },
    { NULL,                                  NULL }
};

/* data per desktop (keep track of all application running on desktop) */
typedef struct 
{
    iml_session_t *auxproxy_session;
    int aux_start[3];
    int wordcount;
}   MyDataPerDesktop;


/* data per session (per application basis) */
typedef struct {
    int status_start;
    int preedit_start;
    int luc_start;
    int conv_on;

    IMText **luc_candidates;
    IMText **luc_labels;

    UTFCHAR *conversion_string;         /* reversed */
    UTFCHAR *preedit_string;            /* underlined */

    int luc_top;
    int luc_nchoices;
    int luc_current_candidate;          /* index of current candidate */
    int luc_type;                       /* Clinet or IM Master */
    int max_candidates;

    UTFCHAR *preedit_buf;               /* preedit */
    IMFeedbackList *preedit_feedback;   /* feedback for preedit */
    int caret_pos;

    IMFeedbackList *luc_fbs_reverse;
    IMFeedbackList *luc_fbs_normal;

    UTFCHAR *commit_buf;	            /* commit */

    UTFCHAR *on_string;
    int cur_im;
        
// START XCIN
    gen_inp_conf_t* cf;
    gen_inp_iccf_t* iccf;
    inpinfo_t* inpinfo;
// END XCIN
    char *tabenc;                        /* encoding that needed in tab */

}   MyDataPerSession;


enum 
{
    SAMPLE_AUX_PANEL = 0,
    SAMPLE_AUX_DICT = 1
};

/*
 * String at the lower left corner of the screen telling user whether the converstation is off
 */
UTFCHAR off_string[] = { 0x82f1, 0x6578, 0 };


/*
 * This method is invoked when htt_server retrieves if_method_t method table.
 */
void if_GetIfInfo(IMArgList args, int num_args)
{
    int i;

	init_objects();
    for(i = 0; i < num_args; i++, args++)
    {
        switch(args->id)
        {
            case IF_VERSION:
                args->value = (IMArgVal) "0.1";
                break;
            case IF_METHOD_TABLE:
                args->value = (IMArgVal) & xcin_method;
                break;
            case IF_LE_NAME:
                args->value = (IMArgVal) & lename;
                break;
            case IF_SUPPORTED_LOCALES:
			    args->value = (IMArgVal) & locales;
                break;
            case IF_SUPPORTED_OBJECTS:
                args->value = (IMArgVal) objects;
                break;
            case IF_NEED_THREAD_LOCK:
                args->value = (IMArgVal) False;
                break;
            default:
                break;
        }
    }
}

/*
 * This method is invoked when the Language Engine is started to use by htt_server
 */
Bool if_xcin_OpenIF( iml_if_t* If )
{
#ifdef  DEBUG
    printf("if_xcin_OpenIF()\n");
    printf("    If=[%x]\n", If);
    printf("    locale=[%s]\n", If->locale);
    printf("    if_name=[%s]\n", If->if_name);
#endif
    
    init_objects();
    return True;
}

/*
 * This method is invoked when htt_server stops to use the Language Engine, 
 * which means there is no client using the Language Engine.
 */
Bool if_xcin_CloseIF( iml_if_t* If )
{
#ifdef  DEBUG
    printf("if_xcin_CloseIF()\n");
    printf("    If=[%x]\n", If);
    printf("    locale=[%s]\n", If->locale);
    printf("    if_name=[%s]\n", If->if_name);
#endif

    free_objects();
    return True;
}

/*
 * This method is invoked when the Language Engine wants to set IF attributes of IM Client.
 */
Bool if_xcin_GetIFValue( iml_if_t* If, IMArgList args, int num_args )
{
#ifdef DEBUG
   printf("*** if_xcin_GetIFValue() ***\n");
   printf("    If=[%x]\n", If);
   printf("    locale=[%s]\n", If->locale);
   printf("    if_name=[%s]\n", If->if_name); 
#endif
    return True;
}

/* 
 * This method is invoked when the IM client wants to set IF attributes.
 */
Bool if_xcin_SetIFValue( iml_if_t* If, IMArgList args, int num_args )
{
#ifdef DEBUG
    printf("*** if_xcin_SetIFValue() ***\n");
    printf("    If=[%x]\n", If);
    printf("    locale=[%s]\n", If->locale);
    printf("    if_name=[%s]\n", If->if_name);
#endif	

    return True;
}

/*
 * This method is invoked during the first connection when the user start to use the desktop
 */
Bool if_xcin_OpenDesktop( iml_desktop_t* desktop, IMArgList args, int num_args )
{
    MyDataPerDesktop *desktop_data = (MyDataPerDesktop *) calloc(1, sizeof(MyDataPerDesktop));
    int i;

#ifdef  DEBUG
    printf("if_xcin_OpenDesktop()\n");
    printf("    If=[%x]\n", desktop->If);
    printf("    desktop=[%x]\n", desktop);
    printf("    locale=[%s]\n", desktop->If->locale);
    printf("    if_name=[%s]\n", desktop->If->if_name);

    printf("    USER:%s\n", desktop->user_name);
    printf("    HOST:%s\n", desktop->host_name);
    printf("    DISPLAY:%s\n", desktop->display_id);

    for (i = 0; i < num_args; i++, args++) 
    {
        switch (args->id) 
        {
            case UI_USER_NAME:
                if (args->value) 
                {
                    printf("    UI_USER_NAME=%s\n", args->value);
                }
                break;
            case UI_HOST_NAME:
                if (args->value) 
                {
                    printf("    UI_HOST_NAME=%s\n", args->value);
                }
                break;
            case UI_DISPLAY_ID:
                if (args->value) 
                {
                    printf("    UI_DISPLAY_ID=%s\n", args->value);
                }
                break;
            case UI_PROTOCOL_TYPE:
                if (args->value) 
                {
                    printf("    UI_PROTOCOL_TYPE=%s\n", args->value);
                }
                break;
            case UI_CLIENT_TYPE:
                if (args->value) 
                {
                    printf("    UI_CLIENT_TYPE=%s\n", args->value);
                }
                break;
            case UI_XSERVER_VENDOR:
                if (args->value) 
                {
                    printf("    UI_XSERVER_VENDOR=%s\n", args->value);
                }
                break;
            case UI_OS_NAME:
                if (args->value) 
                {
                    printf("    UI_OS_NAME=%s\n", args->value);
                }
                break;
            case UI_OS_ARCH:
                if (args->value) 
                {
                    printf("    UI_OS_ARCH=%s\n", args->value);
                }
                break;
            case UI_OS_VERSION:
                if (args->value) 
                {
                    printf("    UI_OS_VERSION=%s\n", args->value);
                }
                break;
            case UI_AUTH_PASSWD:
                if (args->value) 
                {
                    printf("    UI_AUTH_PASSWD=%s\n", args->value);
                }
                break;
            }
    }
#endif

    for ( i = 0; i < MAX_AUX; i++ ) 
    {
        desktop_data->aux_start[i] = False;
    }
    desktop_data->auxproxy_session = 0;
    desktop->specific_data = (void *) desktop_data;
    return True;
}


/*
 * This method is invoked when the user exit from the desktop, which means there is no IM client for the server.
 * An axample is the shutdown of 'httx'
 */
Bool if_xcin_CloseDesktop( iml_desktop_t* desktop )
{
    MyDataPerDesktop *desktop_data = (MyDataPerDesktop *) desktop->specific_data;

#ifdef DEBUG
    printf("*** if_xcin_CloseDesktop()***\n");
    printf("    If=[%x]\n", desktop->If);
    printf("    desktop=[%x]\n", desktop);
    printf("    locale=[%s]\n", desktop->If->locale);
    printf("    if_name=[%s]\n", desktop->If->if_name);

    printf("    USER: %s\n", desktop->user_name);
    printf("    HOST: %s\n", desktop->host_name);
    printf("    DISPLAY: %s\n", desktop->display_id);
#endif

    free(desktop_data);
    return True;
}

/*
 * This method is invoked when a session is created
 */
Bool if_xcin_CreateSC( iml_session_t* s, IMArgList args, int num_args )
{
    iml_desktop_t *desktop = s->desktop;
    int i;

    FILE* fp;
    char filename[256]; 

    MyDataPerSession *p = (MyDataPerSession *) calloc(1, sizeof(MyDataPerSession));
    p->status_start = False;
    p->luc_start = False;
    p->preedit_start = False;
    p->preedit_buf = (UTFCHAR *) calloc(1, sizeof(UTFCHAR) * BUFSIZE);
    p->commit_buf = (UTFCHAR *) calloc(1, sizeof(UTFCHAR) * BUFSIZE);

    p->conversion_string = (UTFCHAR *) calloc(1, sizeof(UTFCHAR) * BUFSIZE);
    p->preedit_string = (UTFCHAR *) calloc(1, sizeof(UTFCHAR) * BUFSIZE);
    p->luc_candidates = NULL;
    p->luc_labels = NULL;
    p->luc_nchoices = 5;
    p->caret_pos = -1;
    p->max_candidates = MAXCANDIDATES - 1;
    p->luc_current_candidate = 0;

    p->preedit_feedback = create_feedback(0, BUFSIZE);
    p->luc_fbs_reverse = create_feedback(0, BUFSIZE);
    p->luc_fbs_normal = create_feedback(0, BUFSIZE);

    /* Start Input Method on first one; */
    p->cur_im = 0;
    p->on_string = imentries[p->cur_im].name;


// START XCIN
    p->cf = calloc(1, sizeof(gen_inp_conf_t));
    p->iccf = calloc(1, sizeof(gen_inp_iccf_t));
    p->inpinfo = calloc(1, sizeof(inpinfo_t));

    p->cf->mode = 66694;                // HARDCODED    TODO: fix in v0.2
    p->cf->inp_ename = imentries[p->cur_im].file; 
    p->tabenc = strdup(TAB_ENC);
    
    sprintf(filename, "%s/%s.tab", TAB_PATH, imentries[p->cur_im].file);
    if ( (fp = fopen( filename, "r" )) == NULL ) 
        printf("*** XCIN: File could not be opened.\n");
    else
    {
        load_syscin();
        gen_inp_xim_init( p->cf, p->iccf, p->inpinfo );
        ccode_info( &(p->cf->ccinfo) );
        loadtab( p->cf, fp, p->tabenc ); 
    }
// END XCIN
    fclose(fp);

    // NOTE: If encounter any preedit problem, have a look here
    for (i = 0; i < BUFSIZE; i++) {
        set_feedback(&p->preedit_feedback[i], IMUnderline);
        set_feedback(&p->luc_fbs_reverse[i], IMReverse);
        set_feedback(&p->luc_fbs_normal[i], IMNormal);
    }

#ifdef  DEBUG
    printf("if_xcin_CreateSC()\n");
    printf("    If=[%x]\n", desktop->If);
    printf("    desktop=[%x]\n", desktop);
    printf("    locale=[%s]\n", desktop->If->locale);
    printf("    if_name=[%s]\n", desktop->If->if_name);

    printf("    BASIC INFO - USER:%s\n", desktop->user_name);
    printf("    BASIC INFO - HOST:%s\n", desktop->host_name);
    printf("    BASIC INFO - DISPLAY:%s\n", desktop->display_id);

    for (i = 0; i < num_args; i++, args++) 
    {
        switch (args->id) 
        {
            case UI_USER_NAME:
                if (args->value) 
                {
                    printf("    UI_USER_NAME=%s\n", args->value);
                }
                break;
            case UI_HOST_NAME:
                if (args->value) 
                {
                    printf("    UI_HOST_NAME=%s\n", args->value);
                }
                break;
            case UI_DISPLAY_ID:
                if (args->value) 
                {
                    printf("    UI_DISPLAY_ID=%s\n", args->value);
                }
                break;
            case UI_PROTOCOL_TYPE:
                if (args->value) 
                {
                    printf("    UI_PROTOCOL_TYPE=%s\n", args->value);
                }
                break;
            case UI_XSERVER_VENDOR:
                if (args->value) 
                {
                    printf("    UI_XSERVER_VENDOR=%s\n", args->value);
                }
                break;
            case UI_CLIENT_TYPE:
                if (args->value) 
                {
                    printf("    UI_CLIENT_TYPE=%s\n", args->value);
                }
                break;
            case UI_OS_NAME:
                if (args->value) 
                {
                    printf("    UI_OS_NAME=%s\n", args->value);
                }
                break;
            case UI_OS_ARCH:
                if (args->value) 
                {
                    printf("    UI_OS_ARCH=%s\n", args->value);
                }
                break;
            case UI_OS_VERSION:
                if (args->value) 
                {
                    printf("    UI_OS_VERSION=%s\n", args->value);
                }
            case UI_AUTH_PASSWD:
                if (args->value) 
                {
                    printf("    UI_AUTH_PASSWD=%s\n", args->value);
                }
                break;
            }
    }
#endif

    s->specific_data = (void *) p;
    return True;
}

/*
 * This method is invoked when the session is destroyed
 * An example is the shudown of an application ie. gedit
 */
Bool if_xcin_DestroySC( iml_session_t* s )
{ 
    MyDataPerSession *p = (MyDataPerSession *) s->specific_data;
    MyDataPerDesktop *desktop_data = (MyDataPerDesktop *) s->desktop->specific_data;

    if ( s == desktop_data->auxproxy_session)
    {
        desktop_data->auxproxy_session = NULL;
    }

    if (p->preedit_buf)
        free((char *) p->preedit_buf);
    if (p->commit_buf)
        free((char *) p->commit_buf);
    if (p->conversion_string)
        free((char *) p->conversion_string);
    if (p->preedit_string)
        free((char *) p->preedit_string);
    if (p->preedit_feedback)
        free((char *) p->preedit_feedback);

    if (p->luc_fbs_reverse)
        free((char *) p->luc_fbs_reverse);
    if (p->luc_fbs_normal)
        free((char *) p->luc_fbs_normal);

    free((char *) p); 

#ifdef DEBUG
    printf("*** if_xcin_DestroySC(s=%x)\n", s);
    printf("    s=[%x]\n", s);
    printf("    If=[%x]\n", s->desktop->If);
    printf("    desktop=[%x]\n", s->desktop);
    printf("    locale=[%s]\n", s->desktop->If->locale);
    printf("    if_name=[%s]\n", s->desktop->If->if_name);

    printf("    USER:%s\n", s->desktop->user_name);
    printf("    HOST:%s\n", s->desktop->host_name);
    printf("    DISPLAY:%s\n", s->desktop->display_id);
#endif

    return True;
}

/*
 * This method is invoked when the Language Engine wants to set SC attributes of IM client
 */
Bool if_xcin_GetSCValue( iml_session_t* s, IMArgList args, int num_args )
{
	return True;
}

/*
 * This method is invoked when the IM client wants to set SC attributes.
 */
Bool if_xcin_SetSCValue( iml_session_t* s, IMArgList args, int num_args )
{
    int i;
    IMArg *p = args;
    MyDataPerDesktop *desktop_data = (MyDataPerDesktop *) s->desktop->specific_data;
    MyDataPerSession *session_data = (MyDataPerSession *) s->specific_data;
                                                                                                
#ifdef  DEBUG
    printf("*** if_xcin_SetSCValue() ***\n");
    printf("    s=[%x]\n", s);
    printf("    If=[%x]\n", s->desktop->If);
    printf("    desktop=[%x]\n", s->desktop);
    printf("    locale=[%s]\n", s->desktop->If->locale);
    printf("    if_name=[%s]\n", s->desktop->If->if_name);
    printf("    USER:%s\n", s->desktop->user_name);
    printf("    HOST:%s\n", s->desktop->host_name);
#endif
                                                                                                                             
    for (i = 0; i < num_args; i++, p++)
    {
        switch (p->id)
        {
            case SC_TRIGGER_ON_NOTIFY:
                conversion_on(s);
                break;
            case SC_TRIGGER_OFF_NOTIFY:
                conversion_off(s);
                break;
            case SC_REALIZE:
                if (!desktop_data->auxproxy_session)
                {
                    desktop_data->auxproxy_session = s;
                    printf("new auxproxy_session %x count=%d\n", desktop_data->auxproxy_session, s->desktop->session_count);
                    if (getenv("DONOTSTART_AUX_AT_SC_REALIZE") == NULL)
                    {
//                        aux_start(desktop_data->auxproxy_session, SAMPLE_AUX_PANEL);
                    }
                }
                break;
            case SC_LOOKUP_LABELTYPE:
                break;
            default:
                break;
        }
    }
                                                                                                                             
    return True;
}

/*
 * This method is invoked when the IM client wants to reset the Input Context.
 */
IMText* if_xcin_ResetSC( iml_session_t* s )
{
    int i;
    iml_inst *lp;
    IMText* p = make_preedit_imtext(s);
    MyDataPerSession *session_data = (MyDataPerSession *) s->specific_data;
                                                                                                                             
#ifdef DEBUG
    printf("*** if_xcin_ResetSC(s=%x) ***\n", s);
    printf("    s=[%x]\n", s);
    printf("    If=[%x]\n", s->desktop->If);
    printf("    desktop=[%x]\n", s->desktop);
    printf("    locale=[%s]\n", s->desktop->If->locale);
    printf("    if_name=[%s]\n", s->desktop->If->if_name);
    printf("    USER:%s\n", s->desktop->user_name);
    printf("    HOST:%s\n", s->desktop->host_name);
    printf("    DISPLAY:%s\n", s->desktop->display_id);
#endif
                                                                                                                             
    /*
     * When you return IMText for commit string, you will need to call iml_make_preedit_erase_inst() here
     */
    lp = s->If->m->iml_make_preedit_erase_inst(s);
    s->If->m->iml_execute(s, &lp);
                                                                                                                             
    /*
     * reset buffer
     */
    for (i = 0; i < BUFSIZE; i++)
    {
        set_feedback(&session_data->preedit_feedback[i], IMUnderline);
    }
    memset(session_data->preedit_buf, 0, sizeof(UTFCHAR) * BUFSIZE);
    memset(session_data->commit_buf, 0, sizeof(UTFCHAR) * BUFSIZE);
    memset(session_data->preedit_string, 0, sizeof(UTFCHAR) * BUFSIZE);
    memset(session_data->conversion_string, 0, sizeof(UTFCHAR) * BUFSIZE);
    session_data->caret_pos = -1;
                                                                                                                             
    /*
     * return committed string
     */
    if ( p -> char_length )
    {
        return p;
    }
                                                                                                                             
    return (IMText *) NULL;
}

/*
 * This method is invoked when the IM client gets the input focus.
 */
void if_xcin_SetSCFocus( iml_session_t* s )
{
                                                                                                                             
#ifdef  DEBUG
    printf("*** if_xcin_SetSCFocus() ***\n");
    printf("    s=[%x]\n", s);
    printf("    If=[%x]\n", s->desktop->If);
    printf("    desktop=[%x]\n", s->desktop);
    printf("    locale=[%s]\n", s->desktop->If->locale);
    printf("    if_name=[%s]\n", s->desktop->If->if_name);
    printf("    USER:%s\n", s->desktop->user_name);
    printf("    HOST:%s\n", s->desktop->host_name);
    printf("    DISPLAY:%s\n", s->desktop->display_id);
#endif
                                                                                                                             
    status_draw(s);
}
                                                                                                                             
/*
 * This method is invoked when the IM client loses the input focus
 */
void if_xcin_UnsetSCFocus( iml_session_t* s )
{
#ifdef DEBUG
    printf("*** if_xcin_UnsetSCFocus()\n");
    printf("    s=[%x]\n", s);
    printf("    If=[%x]\n", s->desktop->If);
    printf("    desktop=[%x]\n", s->desktop);
    printf("    locale=[%s]\n", s->desktop->If->locale);
    printf("    if_name=[%s]\n", s->desktop->If->if_name);
    printf("    USER:%s\n", s->desktop->user_name);
    printf("    HOST:%s\n", s->desktop->host_name);
    printf("    DISPLAY:%s\n", s->desktop->display_id);
#endif
}

/*
 * This method is invoked when any event occurs on the IM client.
 */
void if_xcin_SendEvent( iml_session_t* s, IMInputEvent* ev )
{
#ifdef  DEBUG
    printf("*** if_xcin_SendEvent( ) ***\n");
    printf("    if_xcin_SendEvent s=%x ev=%x\n", s, ev);
#endif
    if (ev)
    {
        if (ev->type == IM_EventKeyList)
        {
            if (receive_keylist(s, (IMKeyListEvent *) ev) == False)
            // changing to True actaully prints out char while u type but no underline
            {
                IMKeyListEvent *keylist = (IMKeyListEvent *) ev;
                iml_inst *lp = s->If->m->iml_make_keypress_inst(s, (IMKeyEventStruct *) keylist->keylist);
                s->If->m->iml_execute(s, &lp);
            }
        }
        // else if for the rest event here :)
    }

}


//===================================================================================//
//=			Helper Function	(non if_xcin function)                                  =//
//===================================================================================//

/*
 * Initialise IM Object
 */ 
void init_objects()
{
    IMObjectDescriptorStruct *l;	
    objects = ( IMObjectDescriptorStruct * ) calloc ( 3, sizeof( IMObjectDescriptorStruct ) );  
    l = objects;	

    l->leid = "xcin";				            /* engine id */
    l->type = IM_DOWNLOADINGOBJECT_JARGUI_TYPE; /* IMObjectType */
    l->name = aux_name_panel;			        /* HRN (Human Readable Name) */
    l->name_length = 7;
    l->domain = "com.redhat";
    l->scope = "xcin";

//    l->path= "/usr/lib/im/locale/zh_TW/xcin/

    l++;
    l->leid = "xcin";				            /* engine id */
    l->type = IM_DOWNLOADINGOBJECT_BINGUI_TYPE; /* IMObjectType */
    l->name = aux_name_panel;			        /* HRN (Human Readable Name) */
    l->name_length = 7;
    l->domain = "com.redhat";
    l->scope = "xcin";
    l->path = "./locale/zh_TW/xcin/aux.so";	    /* path for .so starts from /usr/lib/im/ */
						                        /* in our case: /usr/lib/im/locale/zh_TW/xcin/aux.so */
}

/*
 * Free IM Objects from memory
 */ 
void free_objects()
{
    IMObjectDescriptorStruct *l = objects;

    while ( l-> leid )
    {
        free( l->name );
	    l++;
    };
    free( objects );
    objects = NULL;
}


static void conversion_on( iml_session_t* s )
{
#ifdef DEBUG
    printf("*** conversion_on() ***\n");
#endif

    iml_inst *lp;
    iml_inst *rrv = NULL;

    MyDataPerSession *session_data = (MyDataPerSession *) s->specific_data;
//    int aux_int_data[] = {2};
    session_data->conv_on = True;

    lp = s->If->m->iml_make_start_conversion_inst(s);
    s->If->m->iml_link_inst_tail(&rrv, lp);

    if (session_data->preedit_start == False)
    {
        lp = s->If->m->iml_make_preedit_done_inst(s);
        s->If->m->iml_link_inst_tail(&rrv, lp);
        session_data->preedit_start = True;
    }
    lp = s->If->m->iml_execute(s, &rrv);

    status_draw(s);
}

static void conversion_off( iml_session_t* s ) 
{
#ifdef DEBUG
    printf("*** conversion_off ***\n");
#endif

    iml_inst *lp;
    iml_inst *rrv = NULL;
 
//    int aux_int_data[] = {3};

    MyDataPerSession *session_data = (MyDataPerSession *) s->specific_data;
    session_data->conv_on = False;

    /* if preedit exists, commit the string */
    /* REVIEW: commit seems not a proper behaviour, please review
               commit(s);*/
    
    /* clear the preedit string if the conversion is off */
    session_data->preedit_buf[0] = 0;            

    if (session_data->luc_start == True) 
        lookup_end(s);

    if (session_data->preedit_start == True) 
    {
        lp = s->If->m->iml_make_preedit_done_inst(s);
        s->If->m->iml_link_inst_tail(&rrv, lp);
        session_data->preedit_start = False;
        session_data->caret_pos = -1;
    }
    lp = s->If->m->iml_make_end_conversion_inst(s);
    s->If->m->iml_link_inst_tail(&rrv, lp);

    s->If->m->iml_execute(s, &rrv);

    status_draw(s);	
}


IMFeedbackList* create_feedback( iml_session_t* s, int size )
{
    int i;
    IMFeedbackList *feedback;
    IMFeedback *fb;

#ifdef  DEBUG
    printf("\n### create_feedback( ) ###\n\n");
#endif

    if (s) 
    {
        feedback = (IMFeedbackList *) s->If->m->iml_new(s, sizeof(IMFeedbackList) * size);
    } 
    else 
    {
        feedback = (IMFeedbackList *) calloc(1, sizeof(IMFeedbackList) * size);
    }

    for (i = 0; i < size; i++) 
    {
        IMFeedbackList *fbl = &feedback[i];
        fbl->count_feedbacks = 1;

        if (s) 
        {
            fb = fbl->feedbacks = (IMFeedback *) s->If->m->iml_new(s, sizeof(IMFeedback) * 4);
        } 
        else 
        {
            fb = fbl->feedbacks = (IMFeedback *) calloc(1, sizeof(IMFeedback) * 4);
        }

        memset(fbl->feedbacks, 0, sizeof(IMFeedback) * 4);
    }
    return feedback;
}

void set_feedback( IMFeedbackList* fbl, int normalfeedback )
{
#ifdef  DEBUG
    printf("\n### set_feedback( ) ###\n\n");
#endif

    if (normalfeedback == IMUnderline) 
    {
        set_feedback_private(fbl,
            IMUnderline,
            IM_RGB_COLOR(0, 0, 255),            /* FG: blue */
            IM_RGB_COLOR(255, 255, 255),        /* BG: white */
            IM_RGB_COLOR(0, 0, 255));
    } 
    else if (normalfeedback == IMReverse) 
    {
        set_feedback_private(fbl,
            IMReverse,
            IM_RGB_COLOR(255, 255, 255),        /* FG: white */
            IM_RGB_COLOR(0, 0, 255),            /* BG: blue */
            -1);
    } 
    else 
    {
        set_feedback_private(fbl,
            IMNormal,
            -1,
            -1,
            -1);
    }
}

void set_feedback_private( IMFeedbackList* fbl, int normalfeedback, int fg, int bg, int underline )
{
    int count = 0;
    IMFeedback *fb;

#ifdef  DEBUG
    printf("\n### set_feedback_private( ) ###\n\n");
#endif

    fb = &fbl->feedbacks[count];
    IM_FEEDBACK_TYPE(fb) = IM_DECORATION_FEEDBACK;
    IM_FEEDBACK_VALUE(fb) = normalfeedback;
    count++;

    IM_FEEDBACK_COUNT(fbl) = count;
}

void status_draw( iml_session_t* s )
{
    int len;
    iml_inst *lp;
    iml_inst *rrv = NULL;
    UTFCHAR *str;
    MyDataPerSession *session_data = (MyDataPerSession *) s->specific_data;
    MyDataPerDesktop *desktop_data = (MyDataPerDesktop *) s->desktop->specific_data;
    IMText *p = (IMText *) s->If->m->iml_new(s, sizeof(IMText));

    int aux_int_data_convon[] = {2};
    int aux_int_data_convoff[] = {3};

    memset(p, 0, sizeof(IMText));
    p->encoding = UTF16_CODESET;

#ifdef  DEBUG
    printf("\n### status_draw( ) ###\n\n");
#endif

    if (session_data->conv_on) 
    {
        str = session_data->on_string;
        aux_draw(desktop_data->auxproxy_session, SAMPLE_AUX_PANEL, 1, aux_int_data_convon, 0, NULL);
    } 
    else 
    {
        str = off_string;
        aux_draw(desktop_data->auxproxy_session, SAMPLE_AUX_PANEL, 1, aux_int_data_convoff, 0, NULL);
    }

    len = UTFCHARLen(str);
    p->text.utf_chars = (UTFCHAR *) s->If->m->iml_new(s, sizeof(UTFCHAR) * (len + 1));

    UTFCHARCpy(p->text.utf_chars, str);

    p->char_length = len;

    p->feedback = create_feedback(s, p->char_length);

    if (session_data->status_start == False) 
    {
        lp = s->If->m->iml_make_status_start_inst(s);
        s->If->m->iml_link_inst_tail(&rrv, lp);
        session_data->status_start = True;
    }

    lp = s->If->m->iml_make_status_draw_inst(s, p);
    s->If->m->iml_link_inst_tail(&rrv, lp);

    s->If->m->iml_execute(s, &rrv);
}


void aux_draw( iml_session_t* s, int class_name_id, int count_integers, int *integers, int count_strings, UTFCHAR ** strings )
{
    iml_inst *lp;
    int i;
    int len = 7;
    IMText *lts, *lt;
    IMAuxDrawCallbackStruct *aux;
    MyDataPerDesktop *desktop_data;

#ifdef  DEBUG
    printf("\n### aux_draw( ) ###\n\n");
#endif

    if (!s) return;

    desktop_data = (MyDataPerDesktop *) s->desktop->specific_data;

    if (desktop_data->aux_start[class_name_id] == False) 
    {
        printf("AUX is not started.\n");
        return;
    }

    aux = (IMAuxDrawCallbackStruct *) s->If->m->iml_new(s, sizeof(IMAuxDrawCallbackStruct));
    memset(aux, 0, sizeof(IMAuxDrawCallbackStruct));

    aux->aux_name = class_names[class_name_id];

    aux->count_integer_values = count_integers;
    if (aux->count_integer_values) 
    {
        aux->integer_values = (int *) s->If->m->iml_new(s, sizeof(int) * aux->count_integer_values);
        for (i = 0; i < aux->count_integer_values; i++) 
        {
            aux->integer_values[i] = integers[i];
        }
    }
    aux->count_string_values = count_strings;

    if (aux->count_string_values) 
    {
        aux->string_values = lts = (IMText *)
        s->If->m->iml_new(s, sizeof(IMText) * aux->count_string_values);
        memset(aux->string_values, 0, sizeof(IMText) * aux->count_string_values);
        aux->string_values->encoding = UTF16_CODESET;

        for (i = 0, lt = lts; i < aux->count_string_values; i++, lt++) 
        {
            len = UTFCHARLen(strings[i]);
            lt->text.utf_chars = (UTFCHAR *) s->If->m->iml_new(s, sizeof(UTFCHAR) * (len + 1));
            lt->char_length = len + 1;
            UTFCHARCpy(lt->text.utf_chars, strings[i]);
        }
    }

    lp = s->If->m->iml_make_aux_draw_inst(s, aux);
    s->If->m->iml_execute(s, &lp);
}

/*
 * Determine the lenght of UTFCHAR
 */ 
int UTFCHARLen( UTFCHAR* p )
{
    int i;

#ifdef  DEBUG
    printf("\n### UTFCHARLen( ) ###\n\n");
    printf("    p=[%d]\n", p);
#endif

    for (i = 0; *p; i++)
        p++;

#ifdef  DEBUG
    printf("    Lenght of UTFCHAR i=[%d]\n", i);
#endif
    return i;
}

int UTFCHARCpy( UTFCHAR * dest, UTFCHAR * original )
{
    int i;

#ifdef  DEBUG
    printf("\n### UTFCHARCpy( ) ###\n");
    printf("    dest=[%s]\n", dest);
    printf("    orig=[%s]\n", original);
#endif
    for (i = 0; *original; i++) 
    {
        *dest++ = *original++;
    }
    *dest = 0;
    return i;
}

/*
 * The most important function as it bridges the IIIMF core code to original xcin functions
 * which are located in other files
 */ 
Bool receive_keylist( iml_session_t* s, IMKeyListEvent* keylist) 
{
    int i;
    int index;

    IMKeyEventStruct *k = (IMKeyEventStruct *) keylist->keylist;
    MyDataPerSession *session_data = (MyDataPerSession *) s->specific_data;

    i = UTFCHARLen(session_data->preedit_buf);

#ifdef  DEBUG
    printf("*** receive_keylist( ) ***\n");
    printf("    iml_session_t() keycode=%x,keychar=%c, state=%x\n", k->keyCode, k->keyChar, k->modifier);
    printf("    Preedit Buffer Size[%d]\n", i);
    printf("    Preedit Buffer[%s]\n", session_data->preedit_buf[i]);

    printf("    session_data->inpinfo->ename =[%s]\n", session_data->inpinfo->inp_ename);
    printf("    session_data->inpinfo->n_selkey=[%u] \t session_data->cf->header.n_selkey=[%d]\n", session_data->inpinfo->n_selkey, session_data->cf->header.n_selkey );

#endif

    /* This portion of the code will fill the buffer with the keyboard input */
    if(session_data->luc_start == True)
    {
        /* Used to commit char in the luc window using number keys 1-9 */
        if(k->keyChar >= '0' && k->keyChar <= '9')
        {
            index = k->keyChar - '1';
            
            if(k->keyChar == '0') return True;
            
            /* Catching out of range choices */
            if ( index >= session_data->inpinfo->n_mcch )
            {
                return True;
            }
            if(lookup_commit(s, index)){
                session_data->preedit_buf[0] = 0;
                session_data->iccf->keystroke[0] = 0;
            }
            return True;
        }
        else if(k->modifier & IM_SHIFT_MASK)
        {
            if (k->keyCode == IM_VK_P) 
            {
                lookup_prev(s);
                lookup_draw(s);
                return True;
            } 
            else if (k->keyCode == IM_VK_N) 
            {
                lookup_next(s);
                lookup_draw(s);
                return True;
            }
        }
        else if(k->keyCode == IM_VK_ENTER)
        {
            if(lookup_commit(s, 0)){
                session_data->preedit_buf[0] = 0;
                session_data->iccf->keystroke[0] = 0;
            }
            return True;
        }            
    }

    if (k->modifier & IM_CTRL_MASK && k->modifier & IM_ALT_MASK)
    {
        if(k->keyChar >= '1' && k->keyChar <= '9')
        {
            FILE* fp;
            char filename[256];
                        
            index = k->keyChar - '1';
            if(imentries[index].name[0] != NULL){
                session_data->cur_im = index;
                session_data->cf->inp_ename = imentries[session_data->cur_im].file;
                    
                sprintf(filename, "%s/%s.tab", TAB_PATH, imentries[session_data->cur_im].file);   
                if ( (fp = fopen( filename, "r" )) == NULL )
                    printf("*** XCIN: File could not be opened.\n");
                else
                {
                    gen_inp_xim_init( session_data->cf, session_data->iccf, session_data->inpinfo );
                    ccode_info( &(session_data->cf->ccinfo) );
                    loadtab( session_data->cf, fp, session_data->tabenc ); 
                }
                fclose(fp);
                session_data->on_string = imentries[session_data->cur_im].name;
                status_draw(s);
                
                session_data->preedit_buf[0] = 0;
                session_data->iccf->keystroke[0] = 0;
                preedit_draw(s);
                lookup_end(s);
                
                return True;
            }                
        }
        return False;
    }
    /* change IM using Ctrl-Shift */
    else if ((k->keyCode == IM_VK_CONTROL && k->modifier & IM_SHIFT_MASK) ||
             (k->keyCode == IM_VK_SHIFT   && k->modifier & IM_CTRL_MASK))
    {
        FILE* fp;
        char filename[256];
        
        if(imentries[session_data->cur_im+1].name[0] == NULL)
            session_data->cur_im = 0;
        else
            session_data->cur_im++;
                
        session_data->cf->inp_ename = imentries[session_data->cur_im].file;
        sprintf(filename, "%s/%s.tab", TAB_PATH, imentries[session_data->cur_im].file);   
        if ( (fp = fopen( filename, "r" )) == NULL )
            printf("*** XCIN: File could not be opened.\n");
        else{
            gen_inp_xim_init( session_data->cf, session_data->iccf, session_data->inpinfo );
            ccode_info( &(session_data->cf->ccinfo) );
            loadtab( session_data->cf, fp, session_data->tabenc ); 
        }
        fclose(fp);
        session_data->on_string = imentries[session_data->cur_im].name;
        status_draw(s);

        session_data->preedit_buf[0] = 0;
        session_data->iccf->keystroke[0] = 0;
        preedit_draw(s);
        lookup_end(s);
        
        return True;
    }
    /* used to commit char in the luc window using SPACE */
    else if (k->keyCode == IM_VK_SPACE || k->keyCode == IM_VK_ACCEPT)
    {
        if (i>0)
        {
            if(lookup_commit(s, 0)){
                session_data->preedit_buf[0] = 0;
                session_data->iccf->keystroke[0] = 0; 
            }
            return True;
        }
        else 
        {
            return False;
        }
    }
    else if (k->keyCode == IM_VK_BACK_SPACE)
    {
        if (i > 0) 
        {
            session_data->preedit_buf[i-imentries[session_data->cur_im].n_preedit] = 0;
            session_data->iccf->keystroke[i/imentries[session_data->cur_im].n_preedit-1] = 0;   
            preedit_draw(s);
            session_data->inpinfo = match_keystroke( session_data->cf, session_data->inpinfo, session_data->iccf );
            if ( session_data->inpinfo->n_mcch > 0 )
                lookup_draw(s); 
            else
            {
                lookup_end(s);
            }
            
            /* Prevent the luc window from appearing when preedit buffer is empty */
            if( i-1 == 0)
            {
                lookup_end(s);
                memset(session_data->luc_candidates, 0, MAXCANDIDATES * sizeof(IMText));
            }
            return True;
        } 
        else 
        {
            return False;
        }
    }
    else if (k->keyCode == IM_VK_ESCAPE)
    {
        if (i > 0)
        {
            lookup_end(s);
                        
            session_data->preedit_buf[0] = 0;
            session_data->iccf->keystroke[0] = 0;   
            preedit_draw(s);
            memset(session_data->luc_candidates, 0, MAXCANDIDATES * sizeof(IMText));
           
            return True;
        }
        else
        {
            return False;
        }
    }
    else if (k->keyChar >= '!' && k->keyChar <= '~')
    {
        if(i<INP_CODE_LENGTH){
            /* Following section are assignment to preedit with correct keyname and actual keystroke for search
               immentries[].n_preedit specify how many character are stored into wch_t
               because orginial xcin push two single bytes into one wch_t (which only suppose to store one char)
               we need to seperate out the case of one char and two or more char.
               TODO: redo and make a new struture for keyname
            */ 
            UTFCHAR buf; int j;
            if (imentries[session_data->cur_im].n_preedit == 1) {
                buf = (UTFCHAR)UTF8toUnicode((wchar_t)session_data->cf->header.keyname[key2code(k->keyChar)].wch);
                session_data->preedit_buf[i] = buf;
            } else {
                for(j=0; j<imentries[session_data->cur_im].n_preedit; j++) {
                    buf =             
                      (UTFCHAR)UTF8toUnicode((wchar_t)session_data->cf->header.keyname[key2code(k->keyChar)].s[j]);
                   session_data->preedit_buf[i+j] = buf;
                }
            }
            session_data->preedit_buf[i+imentries[session_data->cur_im].n_preedit] = 0;
            
            session_data->iccf->keystroke[i/imentries[session_data->cur_im].n_preedit] = (char)k->keyChar;
            session_data->iccf->keystroke[i/imentries[session_data->cur_im].n_preedit+1] = 0;
        }
        session_data->inpinfo = match_keystroke( session_data->cf, session_data->inpinfo, session_data->iccf );
                  
        /* This is to fix the label area issue (eg. 1 choice = 1 label instad of 1 choice = 2 label */ 
        if ( session_data->inpinfo->n_mcch > 0 )
            lookup_draw(s); /* XCIN HERE */
        else
        {
            if(session_data->luc_start == True){
                lookup_end(s);
                memset(session_data->luc_candidates, 0, MAXCANDIDATES * sizeof(IMText));
            }
        }
    }
    else
    {
        return False;
    }

#ifdef  DEBUG
    printf("    keycode=%x,keychar=%x, state=%x\n", k->keyCode, k->keyChar, k->modifier);
#endif
    preedit_draw(s);

    return True;
}

void lookup_draw( iml_session_t* s )
{
    IMLookupStartCallbackStruct *start;
    iml_inst *lp;
    wch_t *pmcch;
    MyDataPerSession *session_data = (MyDataPerSession *) s->specific_data;
    int i, k, j=0;

    IMLookupDrawCallbackStruct *draw;
    IMText **candidates;
    IMText **labels;
    int max_len=0;

    
    if (session_data->luc_candidates == NULL) 
    {
        session_data->luc_candidates = (IMText **) s->If->m->iml_new2(s, MAXCANDIDATES * sizeof(IMText));
        memset(session_data->luc_candidates, 0, MAXCANDIDATES * sizeof(IMText));
    }
    candidates = session_data->luc_candidates;

    if (session_data->luc_labels == NULL)
    {
        session_data->luc_labels = (IMText **) s->If->m->iml_new2(s, MAXCANDIDATES * sizeof(IMText));
        memset(session_data->luc_labels, 0, MAXCANDIDATES * sizeof(IMText));
    }
    labels = session_data->luc_labels;

    j = 0; pmcch = session_data->inpinfo->mcch;
    for (i = 0; i < session_data->inpinfo->n_mcch; i++, j++) 
    {
        candidates[j] = (IMText *) s->If->m->iml_new2(s, sizeof(IMText));
        memset(candidates[j], 0, sizeof(IMText));

        candidates[j]->encoding = UTF16_CODESET;
/*      candidates[i]->char_length = UTFCHARLen(input[i]);  */
        candidates[j]->char_length = 1;     /* HARDCODED: fix in v0.2 */

        candidates[j]->text.utf_chars = (UTFCHAR *) s->If->m->iml_new2(s, sizeof(UTFCHAR) * (candidates[j]->char_length + 1));
        memset(candidates[j]->text.utf_chars, 0, sizeof(UTFCHAR));

        candidates[j]->text.utf_chars[0] =(UTFCHAR)  (UTF8toUnicode(pmcch->wch));
        pmcch++;
        candidates[j]->feedback = create_feedback(0, candidates[j]->char_length);

        for (k = 0; k < candidates[j]->char_length; k++)
        {
            set_feedback_private(&candidates[j]->feedback[k],
                IMNormal,
                IM_RGB_COLOR(0, 0, 0),
                IM_RGB_COLOR(192, 192, 192),
                -1);
        }
    }   

    session_data->max_candidates = j-1;

    j =0;
    for (i = '1'; i <= '9'; i++, j++) 
    {
        labels[j] = (IMText *) s->If->m->iml_new2(s, sizeof(IMText));
        memset(labels[j], 0, sizeof(IMText));
        labels[j]->encoding = UTF16_CODESET;
        labels[j]->char_length = 1;
        labels[j]->text.utf_chars = (UTFCHAR *) s->If->m->iml_new2(s, sizeof(UTFCHAR) * (labels[j]->char_length + 1));
        memset(labels[j]->text.utf_chars, 0, sizeof(UTFCHAR));

        labels[j]->text.utf_chars[0] = (UTFCHAR) i;
        labels[j]->feedback = create_feedback(0, labels[j]->char_length);
        for (k = 0; k < labels[j]->char_length; k++) 
        {
            set_feedback_private(&labels[j]->feedback[k],
                IMNormal,
                IM_RGB_COLOR(0, 0, 0),
                IM_RGB_COLOR(192, 192, 192),
                -1);
        }   
    }

    /* lookup start */
    if (session_data->luc_start == False) 
    {
        session_data->luc_top = 0;
        start = (IMLookupStartCallbackStruct *) s->If->m->iml_new(s, sizeof(IMLookupStartCallbackStruct));
        memset(start, 0, sizeof(IMLookupStartCallbackStruct));

        start->whoIsMaster = IMIsMaster;

        start->IMPreference = (LayoutInfo *) s->If->m->iml_new(s, sizeof(LayoutInfo));

        start->IMPreference->choice_per_window = session_data->luc_nchoices;
        start->IMPreference->ncolumns = 1;
        start->IMPreference->nrows = 9;
        start->IMPreference->drawUpDirection = DrawUpHorizontally;
        start->IMPreference->whoOwnsLabel = IMOwnsLabel;

        lp = s->If->m->iml_make_lookup_start_inst(s, start);
        s->If->m->iml_execute(s, &lp);
        session_data->luc_start = True;
    }
    
    /* lookup draw */
    draw = (IMLookupDrawCallbackStruct *) s->If->m->iml_new(s, sizeof(IMLookupDrawCallbackStruct));
    memset(draw, 0, sizeof(IMLookupDrawCallbackStruct));
    draw->index_of_first_candidate = 0;
    draw->index_of_last_candidate = 8;
    draw->n_choices = draw->index_of_last_candidate - draw->index_of_first_candidate + 1;

    draw->title = (IMText *) NULL;

    draw->choices = (IMChoiceObject *) s->If->m->iml_new(s, draw->n_choices * sizeof(IMChoiceObject));
    memset(draw->choices, 0, draw->n_choices * sizeof(IMChoiceObject));

    for (i = 0; i < draw->n_choices; i++) 
    {
        IMText *vt;		/* for value */
        IMText *lt;		/* for label */

        vt = draw->choices[i].value = candidates[i + session_data->luc_top];
        lt = draw->choices[i].label = labels[i];

        if (max_len < vt->char_length)
            max_len = vt->char_length;

        if (i + session_data->luc_top == session_data->max_candidates) 
        {
            draw->index_of_first_candidate = 0;
            draw->index_of_last_candidate = i;
            draw->n_choices = i + 1;
            break;
        }

    }

    draw->max_len = max_len;
    draw->index_of_current_candidate = session_data->luc_current_candidate;

    lp = s->If->m->iml_make_lookup_draw_inst(s, draw);
    s->If->m->iml_execute(s, &lp);
}

void lookup_next(iml_session_t * s)
{
    MyDataPerSession *session_data = (MyDataPerSession *) s->specific_data;
    session_data->luc_top += session_data->luc_nchoices;
    session_data->luc_current_candidate = 0;
    if (session_data->luc_top > session_data->max_candidates) 
    {
        session_data->luc_top = 0;
    }
}

void lookup_prev(iml_session_t * s)
{
    MyDataPerSession *session_data = (MyDataPerSession *) s->specific_data;
    session_data->luc_top -= session_data->luc_nchoices;
    session_data->luc_current_candidate = 0;
    if (session_data->luc_top < 0) 
    {
        int p = session_data->max_candidates % session_data->luc_nchoices;
        session_data->luc_top = session_data->max_candidates - p;
    }
}

Bool lookup_commit(iml_session_t * s, int index)
{
    int i;
    iml_inst *lp;
    iml_inst *rrv = NULL;
    MyDataPerSession *session_data = (MyDataPerSession *) s->specific_data;
    IMText **candidates = session_data->luc_candidates;
    
    index += session_data->luc_top;
    
    if(session_data->luc_candidates == NULL || candidates[index] == NULL)
        return False;

    lp = s->If->m->iml_make_lookup_done_inst(s);
    s->If->m->iml_link_inst_tail(&rrv, lp);
    session_data->luc_current_candidate = 0;

    lp = s->If->m->iml_make_preedit_erase_inst(s);
    s->If->m->iml_link_inst_tail(&rrv, lp);

    lp = s->If->m->iml_make_commit_inst(s, candidates[index]);
    s->If->m->iml_link_inst_tail(&rrv, lp);

    lp = s->If->m->iml_make_lookup_done_inst(s);
    s->If->m->iml_link_inst_tail(&rrv, lp);

    s->If->m->iml_execute(s, &rrv);

    i = UTFCHARLen(session_data->preedit_buf);
    
    /* Empty the buffer so that alphabet will not appear after commiting character */
    session_data->preedit_buf[0] = 0;
    session_data->iccf->keystroke[0] = 0;
    preedit_draw(s);
    
    /* Empty candidates so that next commit won't grab the previous one */
    /* TODO: see if it impacts the preformance */
    memset(session_data->luc_candidates, 0, MAXCANDIDATES * sizeof(IMText));

    session_data->luc_start = False;
}

void lookup_end(iml_session_t *s)
{
    iml_inst *lp;
    MyDataPerSession *session_data = (MyDataPerSession *) s->specific_data;
    
    if (session_data->luc_start == True) {
        lp = s->If->m->iml_make_lookup_done_inst(s);
        s->If->m->iml_execute(s, &lp);
        session_data->luc_start = False;
        return True;
    } else
        return False;
}

/*
 * This function perform the actual output onto the screen
 */
void preedit_draw( iml_session_t* s )
{
    iml_inst *lp;
    iml_inst *rrv = NULL;

#ifdef  DEBUG
    printf("\n### preedit_draw( ) ###\n\n");
#endif

    MyDataPerSession *session_data = (MyDataPerSession *) s->specific_data;
    IMText *p = make_preedit_imtext(s);

    if (session_data->preedit_start == False)
    {
        lp = s->If->m->iml_make_preedit_start_inst(s);
        s->If->m->iml_link_inst_tail(&rrv, lp);
        session_data->preedit_start = True;
    }

    /* the magic line tt makes the preedit works like a charm
     * w/o the line the buffer still can be filled and display after the conversion is off
     */
    lp = s->If->m->iml_make_preedit_draw_inst(s, p);
                
    s->If->m->iml_link_inst_tail(&rrv, lp);
    s->If->m->iml_execute(s, &rrv);
}

IMText* make_preedit_imtext( iml_session_t * s )
{
    int len;
    int i;

#ifdef  DEBUG
    printf("\n### make_preedit_imtext( ) ###\n\n");
#endif

    MyDataPerSession *session_data = (MyDataPerSession *) s->specific_data;
    IMText *p = (IMText *) s->If->m->iml_new(s, sizeof(IMText));
    memset(p, 0, sizeof(IMText));

    p->encoding = UTF16_CODESET;
    len = UTFCHARLen(session_data->preedit_buf);
    p->text.utf_chars = (UTFCHAR *) s->If->m->iml_new(s, sizeof(UTFCHAR) * (len + 1));
    UTFCHARCpy(p->text.utf_chars, session_data->preedit_buf);
    p->char_length = len;

    p->feedback = create_feedback(s, p->char_length);

    for (i = 0; i < p->char_length; i++)
    {
        set_feedback(&p->feedback[i], get_feedback(&session_data->preedit_feedback[i]));
#ifdef DEBUG        
        printf("   p->text.utf_chars[i]: 0x%x[%d %d]\n\n", 
          p->text.utf_chars[i], get_feedback(&p->feedback[i]), get_feedback(&session_data->preedit_feedback[i]));
#endif
    }
    return p;
}

int get_feedback( IMFeedbackList* fbl )
{
#ifdef  DEBUG
    printf("\n### get_feedback( ) ###\n\n"); 
#endif

    /* returns IM_DECORATION_FEEDBACK */
    IMFeedback *fb = &fbl->feedbacks[0];
    return IM_FEEDBACK_VALUE(fb);
}

