/*
 * Copyright (C) 1999 by   XCIN TEAM
 * Copyright (C) 2004 by   Leon Ho <llch@redhat.com>
 *                         Lawrence Lim <llim@redhat.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef _XCIN_H
#define _XCIN_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <iconv.h>

#include <X11/Xmd.h>
#define CARD32BIT CARD32

//#define DEBUG
#define MAX_AUX 2
#define MAXCANDIDATES   40

/*
 * header files for SunIM library
 */
#include "SunIM.h"

/*
 * define If method prototype
 */
Bool if_xcin_OpenIF(iml_if_t *);
Bool if_xcin_CloseIF(iml_if_t *);
Bool if_xcin_GetIFValue(iml_if_t *, IMArgList, int);
Bool if_xcin_SetIFValue(iml_if_t *, IMArgList, int);
Bool if_xcin_OpenDesktop(iml_desktop_t *, IMArgList, int);
Bool if_xcin_CloseDesktop(iml_desktop_t *);
Bool if_xcin_CreateSC(iml_session_t *, IMArgList, int);
Bool if_xcin_DestroySC(iml_session_t *);
Bool if_xcin_GetSCValue(iml_session_t *, IMArgList, int);
Bool if_xcin_SetSCValue(iml_session_t *, IMArgList, int);
IMText* if_xcin_ResetSC(iml_session_t *);
void if_xcin_SetSCFocus(iml_session_t *);
void if_xcin_UnsetSCFocus(iml_session_t *);
void if_xcin_SendEvent(iml_session_t *, IMInputEvent *ev); 

/* Non IF Functions */
void init_objects();
void free_objects();

void conversion_on( iml_session_t* s );
void conversion_off( iml_session_t* s );

IMFeedbackList *create_feedback(iml_session_t * s, int size);

void set_feedback(IMFeedbackList *, int);
void set_feedback_private(IMFeedbackList *, int, int, int, int);
int get_feedback(IMFeedbackList *);

void imstatus_draw( iml_session_t*, UTFCHAR *);  

void lookup_draw( iml_session_t* );  
void lookup_next(iml_session_t *);
void lookup_prev(iml_session_t *);
Bool lookup_commit( iml_session_t*, int );
void lookup_end(iml_session_t *);

void preedit_draw( iml_session_t* );
void status_draw(iml_session_t *);

void aux_draw(iml_session_t *, int, int, int *, int, UTFCHAR **); /* the box at the bottom */

Bool receive_keylist(iml_session_t* , IMKeyListEvent* );

IMText *make_preedit_imtext(iml_session_t *);

int UTFCHARLen(UTFCHAR *);
int UTFCHARCpy(UTFCHAR *, UTFCHAR *);


#ifndef TAB_PATH
#define TAB_PATH "/usr/lib/im/locale/zh_TW/xcin/tab"
#define TAB_ENC "utf-8"
#endif

/***************** Section 1 ********************/

/* ---------------------- xcintool .h ------------------------ */
#ifndef WCH_SIZE
#define WCH_SIZE  4
typedef union
{
    unsigned char s[WCH_SIZE];
    wchar_t wch;   // int type defined in stdlib.h
} wch_t;
#endif
                                                                                                                            
                                                                                                                            
/* -------------------- module.h ---------------------------*/
#define SELECT_KEY_LENGTH 15
#define  N_KEYCODE      50  /* # of valid keys 0-9, a-z, .... */


/* ------------------ /util/cin2tab/gencin.h ---------------*/
#define VERLEN		20
#define ENCLEN		15
#define CIN_ENAME_LENGTH	20
#define CIN_CNAME_LENGTH	20

#define END_KEY_LENGTH     10 /* max # of end keys */

typedef struct
{
    char version[VERLEN];	     /* version number  */
    char encoding[ENCLEN];           /* table encoding  */
    char ename[CIN_ENAME_LENGTH];    /* English name    */
    char cname[CIN_CNAME_LENGTH];    /* Chinese name    */

    wch_t keyname[N_KEYCODE];        /* key name define */ // [wct_t @ xcintool.h] [N_KEYCODE @ module.h]
    char selkey[SELECT_KEY_LENGTH + 1]; /* select keys */
    char endkey[END_KEY_LENGTH + 1]; /* end keys */

    unsigned int n_ichar;        /* # of total chars */
    unsigned int n_icode;        /* # of total keystroke code */
    unsigned char n_keyname;     /* # of keyname needed */
    unsigned char n_selkey;      /* # of select keys */
    unsigned char n_endkey;      /* # of end keys */
    unsigned char n_max_keystroke; /* max # of keystoke for each char */
    char icode_mode;
} cintab_head_t; 

/***************** Section 2 ***********************************/

/* -------------------- xcintool.h ------------------ */
typedef unsigned char ubyte_t;
typedef signed char byte_t;  // xcintool.h

/* -------------------- module.h --------------------- */

#define  N_CCODE_RULE           5       /* # of rules of encoding */
#define  N_ASCII_KEY            95      /* Num of printable ASCII char */

/* For encoding check. */
typedef struct {
    short n;
    ubyte_t begin[N_CCODE_RULE], end[N_CCODE_RULE];
} charcode_t;


typedef struct 
{
    unsigned int total_char;
    ubyte_t n_ch_encoding;
    charcode_t ccode[WCH_SIZE];
} ccode_info_t;    

/***************** Section 3 ***********************************/

/* --------- cin2tab/gencin.h -------------- */
#define  INP_CODE_LENGTH        10      /* max # of keys in a keystroke */

/* --------- gen_inp.h ------------ */
typedef struct
{
    char keystroke[INP_CODE_LENGTH+1];
    wch_t wch;      /* section 1 */
} kremap_t;

/***************** Section 4 ***********************/

/* --------------- cin2tab/gencin.h -------------- */
#define ICODE_MODE1             1       /* one icode & one icode_idx */
#define ICODE_MODE2             2       /* two icode & one icode_idx */

/* For input-code char definition. */
typedef unsigned int    icode_t;
typedef unsigned int  icode_idx_t;
typedef icode_idx_t     ichar_t;

/***************************************************/
typedef struct
{
    char* inp_cname;
    char* inp_ename;
    char* tabfn;
    unsigned int mode;
    cintab_head_t header;   /* section 1 */
    ccode_info_t ccinfo;    /* section 2 */
    ubyte_t modesc;
    char* disable_sel_list;
    int n_kremap;
    kremap_t *kremap;       /* section 3 */

    icode_t* ic1;           /* section 4 */
    icode_t* ic2;
    icode_idx_t* icidx;
    ichar_t* ichar;
} gen_inp_conf_t;

/*=============== gen_inp_iccf_t =====================*/
typedef struct
{
    char keystroke[INP_CODE_LENGTH+1];
    unsigned short mode;
    wch_t *mcch_list;
    int *mkey_list;
    unsigned int n_mcch_list, mcch_hidx, mcch_eidx, n_mkey_list;
} gen_inp_iccf_t;

/*=============== objenc_t ============================*/
typedef struct 
{
    char objname[50];
    char encoding[50];
    char objloadname[100];
} objenc_t;

/*=============== inpinfo_t (module.h) ============================*/ 
/*  Structure for IM in IC.  */
typedef struct {
    int imid;               /* ID of current IM Context */
    void *iccf;             /* Internal data of IM for each IC */
    
    char *inp_cname;            /* IM Chinese name */
    char *inp_ename;            /* IM English name */
    ubyte_t area3_len;          /* Length of area 3 of window (n_char)*/
    ubyte_t zh_ascii;           /* The zh_ascii mode */
    unsigned short xcin_wlen;       /* xcin window length */
//    xmode_t guimode;            /* GUI mode flag */
    
    ubyte_t keystroke_len;      /* # chars of keystroke */
    wch_t *s_keystroke;         /* keystroke printed in area 3 */
    wch_t *suggest_skeystroke;      /* keystroke printed in area 3 */
    
    ubyte_t n_selkey;           /* # of selection keys */
    wch_t *s_selkey;            /* the displayed select keys */
    unsigned short n_mcch;      /* # of chars with the same keystroke */
    wch_t *mcch;            /* multi-char list */
    ubyte_t *mcch_grouping;     /* grouping of mcch list */
    byte_t mcch_pgstate;        /* page state of multi-char */
    
    unsigned short n_lcch;      /* # of composed cch list. */
    wch_t *lcch;            /* composed cch list. */
    unsigned short edit_pos;        /* editing position in lcch list. */
    ubyte_t *lcch_grouping;     /* grouping of lcch list */
    
    wch_t cch_publish;          /* A published cch. */
    char *cch;              /* the string for commit. */
} inpinfo_t;

// ================= misc 
#define MODULE_ID_SIZE 20
#define SYSCIN_VERSION  "20000210"

#define INP_MODE_SELKEYSHIFT 0x00000020 /* selkey shift mode on. */ // gen_inp.h

#define MCCH_ONEPG              0 // module.h

// xcintool.h
#define XCINMSG_NORMAL           0              /* normal               */
#define XCINMSG_WARNING          1              /* warning              */
#define XCINMSG_IWARNING         2              /* internal warnning    */
#define XCINMSG_ERROR           -1              /* error                */
#define XCINMSG_IERROR          -2              /* internal error       */
#define XCINMSG_EMPTY            3              /* pure message printed */




#endif      // end of #define _XCIN_H
